/*
	MarkovGen - A Markov Chain Program
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <random>
#include <argp.h>
#include <cinttypes>
#include <cstdint>
#include <set>
#include <stdexcept>
#include "util.hpp"

namespace
{
	class Arguments
	{
		public:
			Arguments(int argc, char **argv) :
				count(1),
				min(0),
				max(65535)
			{
				argp_parse (&s_argp, argc, argv, 0, 0, this);
			}
			
			uint32_t count;
			uint32_t min;
			uint32_t max;
			std::string prefix;
			std::string suffix;
			
		private:
			error_t Parse(int key, char *arg, struct argp_state *state)
			{
				switch (key)
				{
					case 'c':
						count = StringToInt(arg, state, 1);
						break;
					case 'm':
						min = StringToInt(arg, state, 1);
						break;
					case 'n':
						max = StringToInt(arg, state, 1);
						break;
					case 'p':
						prefix = arg;
						break;
					case 's':
						suffix = arg;
						break;
					default:
						return ARGP_ERR_UNKNOWN;
				}
				return 0;
			}
			
			static error_t parser(int key, char *arg, struct argp_state *state)
			{
				Arguments *args = (Arguments *)state->input;
				return args->Parse(key, arg, state);
			}
			
			static uint32_t StringToInt(const char *arg, const struct argp_state *state, uint32_t min = 0, uint32_t max = 0xFFFFFFFF)
			{
				try
				{
					size_t pos = 0;
					long long n = std::stoll(arg, &pos, 0);
					if ((n >= min) && (n <= max) && !arg[pos]){
						return n;
					}
				}
				catch (std::invalid_argument e){}
				catch (std::out_of_range e){}
				
				argp_error (state, "Argument to %s is not an integer in range: %s", state->argv[state->arg_num], arg);
				throw FormatError("Return from argp_error()");
			}
			
			static const struct argp s_argp;
			static const struct argp_option s_options[];
	};
	
	const struct argp Arguments::s_argp = {Arguments::s_options, Arguments::parser, "A program to write stdin out to randomly sized files", 0, 0, 0};
	const struct argp_option Arguments::s_options[] =
	{
		{"count", 'c', "N", 0, "Produce N samples (default 1)", 0},
		{"min", 'm', "N", 0, "Minimum sample size (default 0)", 0},
		{"max", 'n', "N", 0, "Maximum sample size (default 65535)", 0},
		{"prefix", 'p', "PREFIX", 0, "Prefix the filenames with PREFIX (default: none)", 0},
		{"suffix", 's', "SUFFIX", 0, "Suffix the filenames with SUFFIX (default: none)", 0},
		{0}
	};
	
	void Main(const Arguments &arguments)
	{
		char *mem = new char[arguments.max];
		std::mt19937_64 random;
		File input(stdin);
		
		for (uint32_t i = 1; i <= arguments.count; i++){
			uint32_t n = arguments.min + (random() % (arguments.max - arguments.min));
			input.Read(mem, n);
			
			std::string filename = FormatString("%s%010" PRIu32 "%s", arguments.prefix.c_str(), i, arguments.suffix.c_str());
			File output(filename, true);
			output.Write(mem, n);
		}
		
		delete []mem;
	}
}

int main(int argc, char **argv)
{
	try {
		Arguments arguments(argc, argv);
		Main(arguments);
	}
	catch (Error e) {
		e.OutputMessage();
		return 1;
	}
	return 0;
}
