/*
	MarkovGen - A Markov Chain Program
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cstring>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>

#include "util.hpp"

Error::Error(const char *file, int line, const char *format, ...) :
	m_file(file), m_line(line)
{
	char b[4096];
	
	va_list argp;
	va_start(argp, format);
	vsnprintf(b, sizeof(b), format, argp);
	va_end(argp);
	
	m_message = b;
}

void Error::OutputMessage(bool pos, FILE *f) const
{
	if (pos){
		fprintf(f, "Error in file %s at line %i: %s\n", m_file, m_line, m_message.c_str());
	}
	else {
		fprintf(f, "Error: %s\n", m_message.c_str());
	}
}



struct File::FD
{
	FD(int f, bool o) : fd(f), own(o), count(1){}
	
	~FD()
	{
		if (own){
			close(fd);
		}
	}
	
	int fd;
	bool own;
	uint64_t count;
};

File::~File()
{
	m_fd->count--;
	if (!m_fd->count){
		delete m_fd;
	}
}

File::File(int fd, bool own) : m_fd(new File::FD(fd, own))
{
}

File::File(FILE *f, bool own) : m_fd(new File::FD(fileno(f), own))
{
}

File::File(std::string filename, bool write)
{
	FILE *f = fopen(filename.c_str(), write ? "wb" : "rb");
	if (!f){
		throw FormatError("Could not open file '%s' for %s: %s", filename.c_str(), write ? "writing" : "reading", strerror(errno));
	}
	
	m_fd = new FD(fileno(f), true);
}

File::File(const File &file) : m_fd(file.m_fd)
{
	m_fd->count++;
}

File &File::operator=(const File &file)
{
	if (this != &file){
		m_fd = file.m_fd;
		m_fd->count++;
	}
	
	return *this;
}

void File::Read(void *buffer, size_t length)
{
	char *cbuffer = (char *)buffer;
	while (true){
		ssize_t r = read(m_fd->fd, cbuffer, length);
		if (!r){
			throw EndOfFile();
		}
		
		if (r < 0){
			throw FormatError("Could not read from file: %s", strerror(errno));
		}
		
		length -= r;
		if (!length){
			return;
		}
		cbuffer += r;
	}
}

void File::Write(void *buffer, size_t length)
{
	char *cbuffer = (char *)buffer;
	while (true){
		ssize_t r = write(m_fd->fd, cbuffer, length);
		if (!r){
			throw EndOfFile();
		}
		
		if (r < 0){
			throw FormatError("Could not write to file: %s", strerror(errno));
		}
		
		length -= r;
		if (!length){
			return;
		}
		cbuffer += r;
	}
}

uint64_t File::ReadInt(unsigned int n)
{
	uint8_t b = 0;
	uint64_t r = 0;
	
	for (uint64_t i = 0; i < n; i++){
		if (read(m_fd->fd, &b, 1) != 1){
			throw FormatError("Could not read from file: %s", strerror(errno));
		}
		r |= (uint64_t)b << (i << 3);
	}
	
	return r;
}
	
void File::Skip(uint64_t count)
{
	char b[1024];
	while (count){
		ssize_t r = read(m_fd->fd, b, (count < sizeof(b)) ? count : sizeof(b));
		if (r < 0){
			throw FormatError("Could not blackhole bytes: %s", strerror(errno));
		}
		else if (!r){
			throw EndOfFile();
		}
		count -= r;
	}
}

std::string FormatString(const char *format, ...)
{
	char b[1024];
	int maxlength = 1024;
	
	va_list argp;
	va_start(argp, format);
	int length = vsnprintf(b, maxlength, format, argp);
	va_end(argp);
	
	std::string r = b;
	while (length >= maxlength){
		maxlength = length + 1;
		char *hb = new char[maxlength];
		va_list argp2;
		va_start(argp2, format);
		length = vsnprintf(hb, maxlength, format, argp2);
		va_end(argp2);
		r = hb;
		delete []hb;
	}
	return r;
} 
